module;

#include <format>
#include <fstream>
#include <generator>
#include <iostream>
#include <print>
#include <string>

export module main;

import tile;
import traveler;

std::generator<std::string const &>
lines(std::istream &is) {
	std::string line;
	while (std::getline(is, line)) {
		co_yield line;
	}
	co_return;
}

bool
has_victory(std::vector<std::vector<Tile>> &grid) {
	auto const width = std::ranges::max(grid | std::views::transform([](auto const &row) { return row.size(); }));
	auto const height = grid.size();

	for (int row = 0; row < grid.size(); ++row) {
		for (int col = 0; col < grid[row].size(); ++col) {
			Tile &tile = grid[row][col];

			if (tile.connection(Direction::North) != Direction::North) {
				Traveler traveler{row, col, Direction::South};
				traveler.travel(grid);

				if (traveler.row == row && traveler.col == col) {
					return true;
				}

				if (traveler.direction == Direction::South && height > 7 && traveler.row - row == height) {
					return true;
				}
			}

			if (tile.connection(Direction::West) != Direction::West) {
				Traveler traveler{row, col, Direction::East};
				traveler.travel(grid);

				if (traveler.row == row && traveler.col == col) {
					return true;
				}

				if (traveler.direction == Direction::East && width > 7 && traveler.col - col == width) {
					return true;
				}
			}
		}
	}
	return false;
}

int
entry(std::istream &is) {
	std::vector<std::vector<Tile>> grid;

	for (auto const &line : lines(is)) {
		std::vector<Tile> row;

		for (char ch : line) {
			switch (ch) {
			case ' ': row.push_back(Tile::Blank); break;
			case '+': row.push_back(Tile::Cross); break;
			case '/': row.push_back(Tile::Slash); break;
			case '\\': row.push_back(Tile::Backslash); break;
			default: throw std::runtime_error{std::format("Invalid character '{}'", ch)};
			}
		}

		grid.push_back(row);
	}

	if (has_victory(grid)) {
		std::println("Victory");
	} else {
		std::println("No victory");
	}

	return 0;
}

int
main(int argc, char const **argv) {
	try {
		if (argc < 2) {
			entry(std::cin);
		} else {
			std::ifstream is(argv[1]);
			entry(is);
		}
	} catch (std::exception ex) {
		std::println(std::cerr, "{}", ex.what());
		return 1;
	}
}